/*
** COPYRIGHT (c) 1995, 1999 BY
** DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.
** ALL RIGHTS RESERVED.
**
** THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
** ONLY  IN  ACCORDANCE  OF  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE
** INCLUSION OF THE ABOVE COPYRIGHT NOTICE. THIS SOFTWARE OR  ANY  OTHER
** COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY
** OTHER PERSON.  NO TITLE TO AND  OWNERSHIP OF THE  SOFTWARE IS  HEREBY
** TRANSFERRED.
**
** THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE
** AND  SHOULD  NOT  BE  CONSTRUED  AS A COMMITMENT BY DIGITAL EQUIPMENT
** CORPORATION.
**
** DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS
** SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.
**
** COMPONENT:
** 
**   This module is part of the OpenVMS port of the GnuPG encryption package
** 
** MODULE DESCRIPTION:
** 
**   GnuPG_privileged_nonvax.c is a user-written system service that provides
**   privileged operations needed by the gnupgp program.  Currently:
**
**        - Lock a limited number of pages in memory, effectivly giving
**          the process PSWAPM privilege.
**
**   By putting these operations in a protected shareable image, the
**   main image itself does not need to be installed with privileges.
**   The main image therefore can be linked with traceback and 
**   dynamically loaded modules do not have to be made known images either.
**   
**   When first loaded, the module attempts to translate the exec mode
**   logical name GNUPG_LCKPAG_LIMIT.  If defined, ths logical contains
**   a decimal representation of the maximum number of pages that may be
**   locked on the host by this service.  If not defined, default is
**   1 percent of physical pages on system.
**
**   NOTE: because the change mode is now handled by the image
**   activator, the system service itself must do the argument
**   verification to prevent ACCVIOs and other such problems.
**
**   The associated link procedure, UWSSLNK.COM, will link the object
**   and create the executable. The final step is to copy the
**   executable to SYS$LIBRARY and install it using INSTALL with the
**   /HEADER_RES, /OPEN, /SHARED, and /PROTECTED qualifiers.
**
** AUTHORS:
** 
**   Digital Equipment Corporation (original)
**   
** 
** CREATION DATE: 22-DEC-2007
** 
**  DESIGN ISSUES:
**
**	To be appropriately upwardly-compatible, this module should
**	use the plvdef structure definition from the plvdef.h header
**	file.
**
**	Ideally, all access to processor status (PS) register
**	information should be made using some standard structure
**	definition from the appropriate header file, such as found in
**	pr$r_psdef_bits.  For the sake of convenience, the definition
**	of pr$r_psdef_bits resides in prdef.h, since prdef.h already
**	contains all the bitfield definitions for the "architected"
**	internal processor registers (which, unlike the PS, are
**	accessed by MTPR/MFPR PAL calls).  As part of starlet, prdef
**	now realizes all its definitions using the variant_struct VAX
**	C extension (to declare various members of the union with tag
**	prdef), rendering code using pr$r_psdef_bits as so defined
**	technically non-ANSI-standard.  Furthermore, since the
**	__PAL_READ_PS built-in has type uint64 (not directly
**	compatible with the bit-field definition) it is difficult to
**	avoid causing a store if one decides to use the bitfield
**	designations, which is contrary to one of the goals in this
**	routine (see below).  The routines use defined mask bits to
**	get at the previous mode, but this only works because they're
**	the low-order two bits.
**
**	The OpenVMS VAX MACRO-32 example and the OpenVMS Alpha V1.0
**	BLISS example used "DEFINE_SERVICE" macros in a manner which
**	allowed declaration of kernel- and executive-mode routines in
**	separate statements, mixing of kernel- and executive-mode
**	routines in successive statements and automatic counting of
**	the numbers of both kinds of routines.  The macro capabilities
**	of C are limited, by comparison, so only the counts are
**	calculated (as opposed to hard-coded) here.  Specifically,
**	this example relies on sizeof() constructs to avoid having to
**	hard-code the numbers of routines, but still requires that all
**	executive-mode routines be declared in a single declaration,
**	and that all kernel-mode routines be declared in a single
**	declaration.
**
**	Since this is an example "template" for services which may be
**	executed many times, and since PALcode calls can be somewhat
**	expensive, the services in this routine work to minimize
**	the number of PAL calls made.  Thus, the RD_PS call, needed
**	both (1) to get the previous processor mode as part of
**	checking write access to arguments, and (2) to get the first
**	value to return to the caller, stores its result in the
**	variable ps_temp.  ps_temp should not, and likely will not, be
**	materialized other than in a register (since we don't know
**	where we can safely put it) but there is currently no
**	enforceable compiler guarantee of this, or of similar behavior
**	for a "full-sized" user-written system service.  The link of
**	the image should fail if allocation of storage results in a C
**	RTL call.
**
**
**  COMPILATION, LINKING AND INSTALLATION PROCEDURES:
**   
**  	To compile this routine, use
**
**	$ CC GNUPG_PRIVILEGED_NONVAX + SYS$LIBRARY:SYS$LIB_C /LIBRARY
**	
**	To link this routine, use the following link procedure (also
**	present with more comments as UWSS_LNK.COM):
**	
**      $ LINK  /SHARE=UWSS  /PROTECT  /MAP=UWSS -
**	/SYSEXE  /FULL  /CROSS  /NOTRACE -
**	UWSS,  SYS$INPUT:/OPTIONS
**	GSMATCH=LEQUAL,1,1
**	SYMBOL_VECTOR = ( -
**	                     vms_lock_memory   = PROCEDURE, -
**			     vms_unlock_memory  = PROCEDURE,-
**
**	PSECT=USER_SERVICES,VEC,NOWRT,NOEXE
**	COLLECT=PLV,USER_SERVICES
**
**	To use this routine, copy the executable to SYS$LIBRARY, and
**	install it as follows:
**
**	$ INSTALL REPLACE SYS$LIBRARY:GNUPG_PRIVILEGED_NONVAX /HEADER_RES /OPEN /SHARED /PROTECTED
**
**	 
**  PORTABILITY ISSUES:
**   
**	This program is only intended to run as an example collection
**	of user-written system services, in a privileged shareable
**	image, on OpenVMS Alpha.  The program assumes that OpenVMS Alpha
**	will continue to support ONLY the hierarchical OpenVMS VAX
**	page protection model.  That model is more restrictive than
**	the architected Alpha Alpha PTE model, which, for example, may
**	allow user write access but no kernel write access to a given
**	page.  If the restriction is ever lifted, a probew PAL call
**	for current mode as well as one for previous mode will be
**	required for safe operation.
**
**   
** MODIFICATION HISTORY:
**
*/

/* And now, on with the program... */

#include <types.h>
#include <builtins.h>
#include <chfdef.h>
#include <efndef.h>
#include <far_pointers.h>
#include <ints.h>
#include <lib$routines.h>
#include <plvdef.h>
#include <prdef.h>
#include <prvdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <ssdescrdef.h>
#include <lckdef.h>
#include <lkidef.h>
#include <starlet.h>
#include <syidef.h>
#include <lnmdef.h>
#include <descrip.h>
#ifdef TESTIT
#include <string.h>
#include <stdio.h>
#endif

/*
** To make things easy, the number and name of the user-written system
** services must be known at compile time.  So first define the number
** of services (kernel and exec) and allocate the routine lists, which
** in C-language terms are arrays of pointers to functions returning
** ints.
*/


/* "Forward routine" declarations */
int     rundown_handler();
int     vms_lock_memory();
int	vms_unlock_memory();

/* Kernel and exec routine lists: */

int (*(exec_table[]))() = {
       vms_lock_memory,
       vms_unlock_memory };

/*
** Kernel and exec flags.  Don't allow functions to take 64 bit arguments.
*/
int
    exec_flags [] = {
	0, 0 };
/*
** The next two defines allow the kernel and executive routine counts
** to be filled in automatically after lists have been declared for
** kernel and exec mode.  They must be placed before the PLV
** declaration and initialization, and for this module will be
** functionally equivalent to:
**
** #define KERNEL_ROUTINE_COUNT 0
** #define EXEC_ROUTINE_COUNT 1
** 
*/

#define EXEC_ROUTINE_COUNT sizeof(exec_table)/sizeof(int *)
#define KERNEL_ROUTINE_COUNT 0

/*
** Now build and initialize the PLV structure.  Since the PLV must have
** the VEC psect attribute, and must be the first thing in that psect,
** we use the strict external ref-def model which allows us to put the
** PLV structure in its own psect.  This is like the globaldef
** extension in VAX C, where you can specify in what psect a global
** symbol may be found; unlike globaldef, it allows the declaration
** itself to be ANSI-compliant.  Note that the initialization here
** relies on the change-mode-specific portion (plv$r_cmod_data) of the
** PLV being declared before the portions of the PLV which are specific
** to message vector PLVs (plv$r_msg_data) and system service intercept
** PLVs (plv$r_ssi_data).
**
*/

#ifdef __ALPHA
#pragma extern_model save
#pragma extern_model strict_refdef "USER_SERVICES"
#endif
extern const PLV user_services = {
        PLV$C_TYP_CMOD,         /* type */
        0,                      /* version */
        {
        {KERNEL_ROUTINE_COUNT,	/* # of kernel routines */
        EXEC_ROUTINE_COUNT,     /* # of exec routines */
        0,	                /* kernel routine list */
        exec_table,             /* exec routine list */
        0,		        /* kernel rundown handler */
        rundown_handler,        /* exec rundown handler */
        0,                      /* no RMS dispatcher */
        0,			/* kernel routine flags */
        exec_flags}		/* exec routine flags */
        }
        };
#ifdef __ALPHA
#pragma extern_model restore
#endif
/**************************************************************************/
/* Define global data structures.  The page protection will only allow
 * modifying from exec or kernel mode.
 */
static int control_vec_initialized = 0;	/* status. */
static struct {
    int pagesize;			/* system hardware defined */
    int limit;				/* Global limit of locked pages */
    int used;				/* number used out of limit */
    long int master_lock;		/* lock_id of parent lock */
    int sub_locks;			/* Number of pages we've locked */
    struct dsc$descriptor_s basename_template;
    char sublock_base_name[64];	
} control_vec;

#define MAX_SUB_LOCKS 128
					/* number of sub-locks. */
typedef unsigned long long addr_t;
static struct {
    long int lock_id;			/* child of master lock */
    long int refcount;			/* # of times we locked this page */
    addr_t base_addr;
} page_locks[MAX_SUB_LOCKS];

/**************************************************************************/
/* Support routines.
 */
/*
 * Enable pswapm and syslck, returning previous privilege mask.
 */
static long required_privs[2] = { 
	PRV$M_PSWAPM|PRV$M_SYSLCK,
	0 };
static int disable_privileges ( long *previous )
{
    /*
     * Only disable privileges that user didn't have before.
     */
    long extra_privs[2];
    extra_privs[0] = (required_privs[0] & (~previous[0]));
    extra_privs[1] = (required_privs[1] & (~previous[1]));
    /*
     * Skip $setprv call if nothing to disable
     */
    if ( (extra_privs[0] == 0) && (extra_privs[1] == 0) ) return 1;

    return SYS$SETPRV ( 0, extra_privs, 0, 0 );
}

static int enable_privileges ( long *previous )
{
    int status;
    status = SYS$SETPRV ( 1, required_privs, 0, previous );
    if ( status == SS$_NOTALLPRIV ) {
        disable_privileges ( previous );
	status = SS$_NOSETPRV;
    }
    return status;
}

/*
 * To enforce global limit on locked pages, we get PW mode lock on master, 
 * check current number of sub-locks, allocate a sub-lock for each page, then 
 * lower master lock to let other processes add sub-locks.
 */
static int reserve_begin ( )
{
    int status;
    struct { short length, code; void *buffer; short int *ret_len; } item[2];
    long iosb[2];
    static $DESCRIPTOR(master_lock_name, "GnuPG_LCKPAG_CONTROL");
    struct {
	short status, res;
	long id;
	long valblk[4];
    } lksb;
    /*
     * Initialize master lock if first call.e
     */
    if ( control_vec.master_lock == 0 ) {
	status = SYS$ENQW ( EFN$C_ENF, LCK$K_NLMODE, &lksb, 
		LCK$M_EXPEDITE | LCK$M_SYSTEM,
		&master_lock_name, 0, 0, 0, 0,
		0, 0, 0 );
	if ( (status&1) == 0 ) return status;
	if ( (lksb.status&1) == 0 ) return lksb.status;
	control_vec.master_lock = lksb.id;
    }
    /*
     * Raise lock so only one process at a time is reserving pages.
     */
    lksb.id = control_vec.master_lock;
    status = SYS$ENQW ( EFN$C_ENF, LCK$K_PWMODE, &lksb,
		LCK$M_CONVERT, 0, 0, 0, 0, 0, 0, 0, 0 );
    if ( (status&1) == 0 ) return status;
    if ( (lksb.status&1) == 0 ) return lksb.status;
    /*
     * Update current number of pages consumed from limit
     * by simply getting sub-resource count.
     */
    item[0].code = LKI$_RSBREFCNT;
    item[0].length = sizeof(control_vec.used);
    item[0].buffer = &control_vec.used;
    item[0].ret_len = 0;
    item[1].code = item[1].length = 0;

    status = SYS$GETLKIW ( EFN$C_ENF, &lksb.id, item, iosb, 0, 0, 0 );
    if ( (status&1) == 1 ) status = iosb[0];

    return status;
}
static int reserve_page ( addr_t address )
{
    int status;
    struct { short length, code; void *buffer; short int *ret_len; } item[2];
    long iosb[2];
    short int len;
    static char lock_name[64];
    static $DESCRIPTOR(lock_name_dx, lock_name );
    struct {
	short status, res;
	long id;
	long valblk[4];
    } lksb;
    static char sublock_name[64];
    
    /*
     * Construct name.
     */
    status = SYS$FAO ( &control_vec.basename_template, &len, &lock_name_dx,
	address );
    if ( (status & 1) == 0 ) return status;
    lock_name_dx.dsc$w_length = len;
#ifdef TESTIT
    lock_name[len] = 0;
    printf ( "Lock name for page %llx is %s\n", address, lock_name );
#endif
    /*
     * Get PR mode lock.
     */
    lksb.id = 0;
    status = SYS$ENQW ( EFN$C_ENF, LCK$K_PRMODE, &lksb,
		LCK$M_SYSTEM, &lock_name_dx, 
		control_vec.master_lock, 0, 0, 0, 0, 0, 0 );
#ifdef TESTIT
    printf ( "    status of sublock enq: %d %d\n", status,lksb.status);
#endif
    if ( (status&1) == 0 ) return status;
    if ( (lksb.status&1) == 0 ) return lksb.status;
    /*
     * Add to list of saved pages, assume caller doesn't overflow.
     */
    page_locks[control_vec.sub_locks].lock_id = lksb.id;
    page_locks[control_vec.sub_locks].base_addr = address;
    page_locks[control_vec.sub_locks].refcount = 1;
    control_vec.sub_locks++;

    return status;
}
static int unreserve_page ( lock_ndx )
{
    int status;

    if ( lock_ndx < control_vec.sub_locks ) {
	/*
	 * Release lock and remove from page list.
	 */
	status = SYS$DEQ ( page_locks[lock_ndx], 0, 0, 0 );

	control_vec.sub_locks--;
	if ( control_vec.sub_locks != lock_ndx ) {
	    page_locks[lock_ndx] = page_locks[control_vec.sub_locks];
	}
    } else status = SS$_INVARG;
    return status;
}
static int reserve_end ( )
{
    int status;
    struct { short length, code; void *buffer; short int *ret_len; } item[2];
    long iosb[2];
    struct {
	short status, res;
	long id;
	long valblk[4];
    } lksb;
    /*
     * Lower master lock back to concurrent read, used count is now unreliable.
     */
    lksb.id = control_vec.master_lock;
    status = SYS$ENQW ( EFN$C_ENF, LCK$K_NLMODE, &lksb,
		LCK$M_CONVERT, 0, 0, 0, 0, 0, 0, 0, 0 );
    if ( (status&1) == 0 ) return status;
    if ( (lksb.status&1) == 1 ) status = lksb.status;
    return status;
}

/*
 * The lock_range and unlock range routines are called after
 * privileges are enabled so are able to get system
 * locks call call SYS$LKPAG.  Return value is result
 * of system service call.
 */
static int page_lock_index ( addr_t page_addr )
{
    /*
     * Search list of page locks for matching base address.
     * Return -1 if not found.
     */
    int j;
    for ( j = 0; j < control_vec.sub_locks; j++ ) {
       if ( page_locks[j].base_addr == page_addr ) return j;
    }
    return -1;		/* Not found */
}

static int unlock_range ( addr_t *page_addr, int page_count )
{
    int status, i, ndx;
    /*
     * Search for the page_lock for each page in range.
     */
    for ( i = 0; i < page_count; i++ ) {
	ndx = page_lock_index ( page_addr[i] );
#ifdef TESTIT
printf ( "Unlock_range: page[%d] %llx has index %d, ref %d\n", i,
page_addr[i], ndx, (ndx < 0) ? 0 : page_locks[ndx].refcount );
#endif
	if ( ndx >= 0 ) {
	   /*
	    * Lock found, decrement reference count and
	    * delete lock when it goes to zero.
	    */
	   page_locks[ndx].refcount -= 1;
	   if ( page_locks[ndx].refcount <= 0 ) {
		addr_t d1,d2;
		status = SYS$ULKPAG_64 ( page_locks[ndx].base_addr, 
		    control_vec.pagesize, 0, &d1, &d2 );
		if ( (status&1) == 0 ) {
#ifdef TESTIT
		    printf ( "Unlock_range failed on page %llx\n",
			page_locks[ndx].base_addr );
#endif
		}
		unreserve_page ( ndx );
	   }
	}
    }

    return 1;
}

static int lock_range ( addr_t *page_addr, int page_count )
{
    int status, i, ndx;
    addr_t locked_addr, locked_len;
    /*
     * Get global lock to allocate pages.
     */
    status = reserve_begin ( );
    if ( (status&1) == 0 ) return status;
#ifdef TESTIT
	printf ( "locking range of %d, used=%d/%d\n", page_count,
control_vec.used, control_vec.limit );
#endif
    if ( (page_count + control_vec.used) <= control_vec.limit ) {
	/*
	 * Get sublock for each page we want to lock or bump
	 * reference count.
	 */
	for ( i = 0; i < page_count; i++ ) {
	    ndx = page_lock_index ( page_addr[i] );
	    if ( ndx < 0 ) {
	        status = reserve_page ( page_addr[i] );
#ifdef TESTIT
	        printf ( "  sublock %llx result: %d\n", page_addr[i], status );
#endif
	    } else {
#ifdef TESTIT
		printf ( "  sublock %llx refcount increment...\n", page_addr[i]);
#endif
		page_locks[ndx].refcount++;
	    }
 	}
	/*
	 * Lock pages in memory.
	 */
	status = SYS$LCKPAG_64 ( page_addr[0], 
		control_vec.pagesize*page_count, PSL$C_USER, 
		&locked_addr, &locked_len );

	if ( (status&1) == 0 ) {
	    unlock_range ( page_addr, page_count );
	}
    } else {
	status = SS$_LCKPAGFUL;
    }
    /*
     * Ensure we release the allocation lock.
     */
    reserve_end();

    return status;
}



/*
 * Translate the logical GNUPG_LCKPAG_LIMIT and initialize
 * the control vector.
 */
static int verify_caller()
{
    int status, length, i, j, attr;
    struct { short length, code; void *buffer; short int *ret_len; } item[8];
    long iosb[2], memsize, pid, mask;
    char username[12], *src;
    char equiv[256];		/* buffer for result */
    short int equiv_len[2];
    static char logical_name[32];
    static $DESCRIPTOR(logical_name_dx,logical_name);
    static $DESCRIPTOR(name_table, "LNM$FILE_DEV" );
    static $DESCRIPTOR(template_ctlstr, "!XL_!!XL");
    static $DESCRIPTOR(basename_template, "");
    /*
     * Default to basic access.
     */
    control_vec_initialized = 1;
    control_vec.pagesize = 8192;
    control_vec.limit = -1;		/* use default */
    control_vec.master_lock = 0;
    control_vec.sub_locks = 0;
    /*
     * Query system for page size of physical pages available.
     */
    item[0].code = SYI$_PAGE_SIZE;
    item[0].length = sizeof(control_vec.pagesize);
    item[0].buffer = &control_vec.pagesize;
    item[0].ret_len = 0;
    item[1].code = SYI$_MEMSIZE;
    item[1].length = sizeof(memsize);
    item[1].buffer = &memsize;
    item[1].ret_len = 0;
    item[2].length = item[2].code = 0;

    status = SYS$GETSYIW ( EFN$C_ENF, 0, 0, item, iosb, 0, 0 );
    if ( (status&1) == 0 ) return status;
    if ( (iosb[0] & 1) == 0 ) return iosb[0]*0x0ffff;
    /*
     * Construct base name for sub-locks from current process ID
     */
    item[0].length = item[0].code = 0;
    pid = 0;
    status = SYS$GETJPIW ( EFN$C_ENF, &pid, 0, item, iosb, 0, 0 );
    if ( (status&1) == 0 ) return status;
    if ( (iosb[0] & 1) == 0 ) return status;
    control_vec.basename_template = template_ctlstr;
    control_vec.basename_template.dsc$w_length = 
	sizeof(control_vec.sublock_base_name)-1;
    control_vec.basename_template.dsc$a_pointer =
	control_vec.sublock_base_name;
    status = SYS$FAO ( &template_ctlstr,  &equiv_len[1],
	&control_vec.basename_template, pid );
    if ( (status&1) == 0 ) return status;
    control_vec.basename_template.dsc$w_length = equiv_len[1];
    control_vec.sublock_base_name[equiv_len[1]] = 0;
    /*
     * Make descriptor for logical name.
     */
    src = "GNUPG_LCKPAG_LIMIT";
    logical_name_dx.dsc$w_length = 18;
    logical_name_dx.dsc$a_pointer = src;
    /*
     * Translate the logical, note translation will be at
     * exec mode so user's can override it.
     */
    item[0].code = LNM$_STRING;
    item[0].length = sizeof(equiv)-1;
    item[0].buffer = equiv;
    item[0].ret_len = &equiv_len[0];
    item[1].code = item[1].length = 0;		/* terminate list */
    attr = LNM$M_CASE_BLIND;
    status = SYS$TRNLNM ( &attr, &name_table, &logical_name_dx, 0, item );
    /*
     * Convert the decimal strings into integers.
     */
    if ( (status&1) == 1 ) {
	    int value;
	    char *s;
	    /* Scan string and decode, we can't use RTL function */
	    s = equiv;
	    value = 0;
	    for ( j = 0; j < equiv_len[0]; j++ ) {
		if ( (s[j] >= '0') && (s[j] <= '9') ) {
		    value = (value*10) + (s[j] - '0');
		} else break;
		value = value&0x0ffff;
	    }
	    control_vec.limit = value;
    } else {
	/*
	 * Logical name not defined, make limit 1% of pages.
	 */
	control_vec.limit = memsize / 100;
	status = 1;
    }
    return status;
}
/**************************************************************************
** And now, the service routines.  The routines in this example
** execute no instructions which would require privilege to complete,
** but they give different results when run privileged than they would
** if run from a non-privileged shareable image.  Each routine will
** expects to be passed a pointer to an array of 2 longwords.  The
** first longword will be set by the service routine to contain the
** low 32-bits of the processor status register, from which a caller
** can verify that the service routine is executing in the proper mode
** (kernel or exec).  The second longword is then set to a unique
** integer (1 for first_service, and so on up to 4) to verify that the
** correct service routine has been called.
**
** Each routine will verify that it can write to the longword array in
** the previous mode, since we're running in one of the inner modes.
** If a bogus address were passed (either an inaccessible address,
** causing an access violation, or an address accessible in an inner
** mode that shouldn't be accessible in user mode), then without this
** check, either the process would get blown away (exec mode) or the
** system would crash (in kernel mode, right away if an accvio or
** later if system space were corrupted).  Clearly, neither failure
** mode is desirable.  Note that the use of the single PROBEW PAL call
** for previous mode is enough because only a single contiguous pair
** of longword data (equivalent to a single quadword datum) will be
** stored.  For storage of larger amounts of data which may cross more
** than one page boundary, more stringent checking is required.
**
** One final note on the service routines.  As a security precaution,
** protected shareable images aren't allowed to call other shareable
** images, unless they too are installed protected.  Watch your calls
** (including implicit ones) to external routines!  They must either
** be linked (from object files or libraries) into your image, or be
** in other protected images.  If you fail to heed this warning, the
** image activator will 'kindly' remind you with the error:
**
** SYSTEM-F-NOSHRIMG, privileged shareable image cannot have outbound calls
*/
/*
** The vms_lock_memory function locks the pages
**
*/

int vms_lock_memory ( void *addr, size_t length )
{
    register unsigned int ps_temp;  /* Assumed to take no real storage*/
    long prev_privs[2];
    addr_t buffer_start, buffer_end;
    addr_t page_addr[MAX_SUB_LOCKS];
    int page_count, i, status;
    /*
     * Validate caller's arguments and permission to use this service.
     */
    if ( !control_vec_initialized ) verify_caller();
    if ( (control_vec.limit) <= 0 ) return SS$_NOSYSPRV;

    ps_temp = __PAL_RD_PS();	/* Get the value of the proc status reg*/
    if ( __PAL_PROBEW(addr, length,
		ps_temp & PR$M_PS_PRVMOD) == 0 /* previous mode */
       ) return SS$_ACCVIO;		/* nope, sorry */
    /*
     * Determine the base addresses of the pages that must be locked
     * to cover the range specified by the caller.
     */
    buffer_start = (addr_t) addr;
    buffer_end = buffer_start + length - 1;
    page_addr[0] = buffer_start & ~(control_vec.pagesize-1);
#ifdef TESTIT
    printf ( "User's buffer: %llx - %llx, page0: %llx/%d\n", buffer_start, buffer_end, page_addr[0], control_vec.pagesize );
#endif
    for ( i = 1; i < MAX_SUB_LOCKS; i++ ) {
#ifdef TESTIT
    printf ( "page[%d]: %llx\n", i-1, page_addr[i-1] );
#endif
	if (buffer_end < (page_addr[i-1] + control_vec.pagesize)) break;
	page_addr[i] = page_addr[i-1] + control_vec.pagesize;
    }
    page_count = i;
    if ( (page_count+control_vec.sub_locks) > MAX_SUB_LOCKS ) {
	/* Exceeded per-process limit on number of pages */
	return SS$_LCKPAGFUL;
    }
    /*
     * Lock the pages.
     */
    status = enable_privileges ( prev_privs );
    if ( (status&1) == 1 ) {
#ifdef TESTIT
	printf ( "Pages to lock: %d\n", page_count );
#endif
	status = lock_range ( page_addr, page_count );
        disable_privileges ( prev_privs );
    }
    return status;
}


int vms_unlock_memory ( void *addr, size_t length )
{
    register unsigned int ps_temp;  /* Assumed to take no real storage*/
    long prev_privs[2];
    addr_t buffer_start, buffer_end;
    addr_t page_addr[MAX_SUB_LOCKS];
    int page_count, i, status;
    /*
     * Validate caller's arguments and permission to use this service.
     */
    if ( !control_vec_initialized ) verify_caller();
    if ( (control_vec.limit) <= 0 ) return SS$_NOSYSPRV;

    ps_temp = __PAL_RD_PS();	/* Get the value of the proc status reg*/
    if ( __PAL_PROBEW(addr, length,
		ps_temp & PR$M_PS_PRVMOD) == 0 /* previous mode */
       ) return SS$_ACCVIO;		/* nope, sorry */
    /*
     * Determine the base addresses of the pages that must be locked
     * to cover the range specified by the caller.
     */
    buffer_start = (addr_t) addr;
    buffer_end = buffer_start + length - 1;
    page_addr[0] = buffer_start & ~(control_vec.pagesize-1);
    for ( i = 1; i < MAX_SUB_LOCKS; i++ ) {
	if (buffer_end < (page_addr[i-1] + control_vec.pagesize)) break;
	page_addr[i] = page_addr[i-1] + control_vec.pagesize;
    }
    page_count = i;
    if ( (page_count) > MAX_SUB_LOCKS ) {
	/* Exceeded per-process limit on number of pages */
	return SS$_LCKPAGFUL;
    }
    /*
     * Unlock the pages.
     */
    status = enable_privileges ( prev_privs );
    if ( (status&1) == 1 ) {
	status = unlock_range ( page_addr, page_count );
        disable_privileges ( prev_privs );
    }
    return status;
}

/*
**
** This routine is provided simply as a placeholder for a
** real rundown handler.  A user-written rundown handler
** should not invoke any RMS services or RTL routines, and
** must not signal any exceptions.  User-written rundown
** handlers can invoke most system services except those
** that use RMS (i.e.  $PUTMSG).
*/
int rundown_handler()
{
    int i, status;
    /*
     * Unlock the exec mode locks.
     */
    if ( !control_vec_initialized ) return SS$_NORMAL;
    if ( !control_vec.master_lock ) return SS$_NORMAL;

    if ( control_vec.sub_locks ) {
	status = SYS$DEQ ( control_vec.master_lock, 0, 0, LCK$M_DEQALL );
	control_vec.sub_locks = 0;
    }
    status = SYS$DEQ ( control_vec.master_lock, 0, 0, 0 );
    return SS$_NORMAL;			/* Indicate success */
}

#ifdef TESTIT
/*
 * Test program, calls functions in user mode for testing
 * (Only call printf in user mode).
 */
#include <stdio.h>
#include <unistd.h>
int main ( int argc, char **argv )
{
    int status;
    static char pool[100000];

    status = verify_caller ();
    printf ( "verify_caller result: %d ctlvec: %d %d %d\n", status,
	control_vec.pagesize, control_vec.limit, control_vec.used );
    /*
     * Lock some pages twice to verify reference counting works.
     */
    status = vms_lock_memory ( pool, 32000 );
    printf ( "Status of lock memory: %d, addr: %x\n", status, pool );
    status = vms_lock_memory ( pool+9000, 7000 );
    printf ( "Status of second lock: %d\n", status );

    if ( status & 1 ) {
	/*
	 * Sleep so we can verify "L" pages with show process/continuous, then
	 * unlock the big range and verify only some pages unlocked.
	 */
	sleep ( 30 );
	status = vms_unlock_memory ( pool, 32000 );
	printf ( "status of unlock: %d\n", status );
	sleep ( 30 );
	status = vms_unlock_memory ( pool+9000, 7000 );
	printf  ( "Status of 2nd unlock: %d\n", status );
	sleep ( 30 );
    }
    return status;
}
#endif
